import exec_helpers
import multiprocessing
import os
import yaml
from abc import abstractmethod

from concurrent import futures
from retry import retry
from paramiko.ssh_exception import NoValidConnectionsError

from si_tests import logger
from si_tests import settings
from si_tests.utils import utils, templates
from si_tests.deployments.utils import file_utils

LOG = logger.logger


class WorkloadBase:

    @abstractmethod
    def deploy(self):
        pass

    @abstractmethod
    def delete(self):
        pass

    @abstractmethod
    def check(self):
        pass

    @abstractmethod
    def collect_logs(self):
        pass


class PerNodeWorkload(WorkloadBase):
    def __init__(self, ostcm):
        self.ostcm = ostcm
        self.os_manager = self.ostcm.os_manager
        self.stack_prefix = "pernode-test"
        self._stacks_data = {}
        self.template_path = file_utils.join(
            os.path.dirname(os.path.abspath(__file__)),
            "../../tests/lcm/templates/heat_stack_host.yaml",
        )
        self.ssh_keys = utils.generate_keys()

    def _ensure_hugepages_flavor(self, hostname):
        LOG.info("Create a flavor with huge pages")
        flavor_name = "m1.tiny.dpdk-" + hostname
        flavors = self.ostcm.flavor.list(["--all", "--long"])
        current_flavor = [
            flavor for flavor in flavors if flavor.get("Name") == flavor_name
        ]
        if current_flavor:
            LOG.info(f"Flavor with name: {flavor_name} already exists")
        else:
            LOG.info(f"Creating new flavor with name: {flavor_name}")
            flavor_params = [
                "--ram",
                "1024",
                "--disk",
                "5",
                "--vcpus",
                "2",
                "--property",
                "hw:mem_page_size=2048",
                flavor_name,
            ]
            self.ostcm.flavor.create(flavor_params)
        return flavor_name

    def _ensure_stack_exists(self, stack_name, stack_data):
        if not self.ostcm.stack.exists(stack_name):
            LOG.info(f"Stack {stack_name} was not found, creating")
            hostname = stack_data["hostname"]
            node_labels = stack_data["node_labels"]
            az = stack_data["az"]
            if node_labels.get("openstack-compute-node-dpdk", "") == "enabled":
                flavor_name = self._ensure_hugepages_flavor(hostname)
                stack_params = {
                    "host_name": az + ":" + hostname,
                    "key_name": "key-" + hostname,
                    "image_name_1": self.ostcm.cirros_image_name,
                    "flavor1": flavor_name,
                }
            elif node_labels.get("openstack-compute-node", "") == "enabled":
                stack_params = {
                    "host_name": az + ":" + hostname,
                    "key_name": "key-" + hostname,
                    "image_name_1": self.ostcm.cirros_image_name,
                }
            self.ostcm.create_stack(
                '', stack_name, self.template_path, stack_params=stack_params, finalizer=False
            )
        else:
            LOG.info(f"Stack {stack_name} exists, skipping creation")

    def _verify_stack_workload(self, stack_name):
        ssh_access_key = self.ostcm.stack.output_show(
            [stack_name, "heat_key", "--column", "output_value"]
        )["output_value"]
        pkey = utils.get_rsa_key(ssh_access_key)

        @retry((TimeoutError, NoValidConnectionsError, ConnectionResetError), tries=5, delay=30)
        def _ssh_instance(ip):
            LOG.info(
                f"Attempt to connect to VM with ip: {ip} in stack {stack_name}"
            )
            auth = exec_helpers.SSHAuth(username="cirros", password="", key=pkey)
            ssh = exec_helpers.SSHClient(host=ip, port=22, auth=auth)
            ssh.logger.addHandler(logger.console)
            ssh.sudo_mode = True
            return ssh

        servers = [
            x["resource_name"]
            for x in self.ostcm.stack.resource_list(
                [stack_name, "--filter", "type=OS::Nova::Server"], yaml_output=True
            )
        ]
        assert servers, f"No servers existed for stack {stack_name}"

        def _get_fip_only(server):
            server_info = self.ostcm.stack.resource_show([stack_name, server])
            network = list(server_info.get("attributes").get("addresses").keys())[0]
            for address in server_info.get("attributes").get("addresses").get(network):
                addr_type = address["OS-EXT-IPS:type"]
                if addr_type == "floating":
                    LOG.info(
                        "Verify instance {0} with floating ip {1}".format(
                            server_info.get("attributes").get(
                                "OS-EXT-SRV-ATTR:hostname"
                            ),
                            address["addr"],
                        )
                    )
                    return address["addr"]

        failed_results = {}
        ips = []
        with futures.ThreadPoolExecutor() as executor:
            jobs = {
                executor.submit(_get_fip_only, server): server for server in servers
            }
            for future in futures.as_completed(jobs):
                server = jobs[future]
                try:
                    data = future.result()
                    ips.append(data)
                except Exception as e:
                    raise ValueError(
                        f"Getting server {server} IP finished with exception: {e}"
                    )
        for ip in ips:
            ssh = _ssh_instance(ip)
            expected_string = utils.gen_random_string(10)
            ssh.check_call(
                "echo {expected_string} > test.txt".format(
                    expected_string=expected_string
                )
            )
            res = ssh.check_call("cat test.txt")
            if res.stdout_brief != expected_string:
                failed_results[ip] = {
                    "actual": res.stdout_brief,
                    "expected": expected_string,
                }
                LOG.error(f"Instance with IP {ip} failed to check")

        assert (
            not failed_results
        ), f"Some servers return wrong response in stack {stack_name}:\n{yaml.dump(failed_results)}"
        LOG.info("All instances checks passed")

    def _execute_stacks_tasks(self, tasks_map):
        process_data = {}
        for stack_name, task in tasks_map.items():
            process_data[stack_name] = {
                "process": multiprocessing.Process(
                    target=task["target"], args=task["args"], name=task["name"]
                ),
                "exit_code": None,
            }
            process_data[stack_name]["process"].start()
        # Waiting for all processes finished
        for k, v in process_data.items():
            v["process"].join()
            v["exit_code"] = v["process"].exitcode

        # Check processes exit codes
        assert all([v["exit_code"] == 0 for k, v in process_data.items()]), (
            "Next tasks are FAILED:\n",
            "\n".join(
                [
                    f"{v['process'].name} for stack {k}"
                    for k, v in process_data.items()
                    if v["exit_code"] != 0
                ]
            ),
        )

    @property
    def stacks_data(self):
        if not self._stacks_data:
            for hs in self.ostcm.host.list([]):
                if hs.get("Service") == "compute" and "ironic" not in hs.get(
                    "Host Name"
                ):
                    hostname = hs["Host Name"]
                    node_labels = (
                        self.os_manager.api.nodes.get(name=hostname)
                        .read()
                        .to_dict()
                        .get("metadata", {})
                        .get("labels", {})
                    )
                    if (
                        node_labels.get("openstack-compute-node", "") == "enabled"
                        or node_labels.get("openstack-compute-node-dpdk", "")
                        == "enabled"
                    ):
                        stack_name = self.stack_prefix + hostname
                        self._stacks_data[stack_name] = {
                            "hostname": hostname,
                            "node_labels": node_labels,
                            "az": hs["Zone"],
                        }
        return self._stacks_data

    def deploy(self):
        tasks_map = {}
        if settings.OPENSTACK_ENCRYPTED_VOLUME:
            assert self.os_manager.is_encrypted_volume_type_described(), \
                "Encrypted volume type WAS NOT found in osdpl, " \
                "but settings.OPENSTACK_ENCRYPTED_VOLUME is enabled"
        LOG.info("Unset quotas for workloads")
        self.ostcm.quota.set(
            [
                "--volumes",
                "-1",
                "--instances",
                "-1",
                "--cores",
                "-1",
                "--routers",
                "-1",
                settings.MOSK_WORKLOAD_PROJECT_NAME
            ]
        )
        template = templates.render_template(
            self.template_path, options={"ssh_public_key": self.ssh_keys["public"],
                                         "ssh_private_key": self.ssh_keys["private"]})
        heat_template = yaml.safe_load(template)
        with open(self.template_path, 'w') as f:
            f.write(yaml.dump(heat_template))
        for stack_name, data in self.stacks_data.items():
            tasks_map[stack_name] = {
                "target": self._ensure_stack_exists,
                "args": (stack_name, data),
                "name": "ensure_stack_exists",
            }
        self._execute_stacks_tasks(tasks_map)

    def check(self):
        LOG.info("Checking Per node stacks workloads")
        tasks_map = {}
        for stack_name in self.stacks_data.keys():
            tasks_map[stack_name] = {
                "target": self.ostcm.check_stack,
                "args": (stack_name,),
                "name": "check_stack",
            }
        self._execute_stacks_tasks(tasks_map)

        if not self.os_manager.is_tf_enabled():
            # Create and run processes for verifying stacks workloads
            for stack_name in self.stacks_data.keys():
                tasks_map[stack_name] = {
                    "target": self._verify_stack_workload,
                    "args": (stack_name,),
                    "name": "verify_stack",
                }
            self._execute_stacks_tasks(tasks_map)

    def delete(self):
        LOG.info("Deleting Per node stacks")
        tasks_map = {}
        for stack_name in self.stacks_data.keys():
            LOG.info(f"Removing stack {stack_name}")
            tasks_map[stack_name] = {
                "target": self.ostcm.stack.delete,
                "args": ([stack_name, "-y", "--wait"],),
                "name": "delete_stack",
            }
        self._execute_stacks_tasks(tasks_map)

    def collect_logs(self):
        """TODO"""
        pass
