import pytest

from si_tests import logger
from si_tests.utils import waiters

LOG = logger.logger


class TestTFDisableAnalytics(object):
    """Check disable TF analytics feature"""

    @pytest.fixture(scope='class', autouse=True)
    def restore_tfoperator_cr(self, tf_manager):
        if tf_manager.is_analytics_disabled():
            pytest.skip("TF Analytics is disabled. Test should be run against "
                        "environment with enabled Analytics")
        LOG.info("Backup TFOperator specs")
        tfoperator = tf_manager.tfoperator()
        spec_backup = tfoperator.data['spec']
        yield
        LOG.info("Restore TFOperator specs")
        tfoperator.patch({"spec": spec_backup})
        waiters.wait(
            lambda: len(tf_manager.api.pods.list_starts_with(
                'tf-analytics')) != 0,
            timeout=300, interval=10
        )
        tf_manager.wait_tf_controllers_healthy(timeout=240)
        tf_manager.wait_tfoperator_healthy(timeout=180)

    def test_disable_analytics(self, show_step, tf_manager):
        """ Disable TF analytics component

            Scenario:
                1. Disable TF Analytics
                2. Check analytics components
                3. Check cassandra analytics component
                4. Check zookeeper analytics component
                5. Check kafka component
                6. Delete pvc for disabled components
                7. Restart vRouters
                8. Wait TFOperator status
                9. Run TFTests
        """
        show_step(1)
        tf_manager.disable_analytics(True)

        show_step(2)
        waiters.wait(
            lambda: len(tf_manager.api.pods.list_starts_with(
                'tf-analytics')) == 0,
            timeout=120, interval=10
        )

        show_step(3)
        waiters.wait(
            lambda: len(tf_manager.api.pods.list_starts_with(
                'tf-cassandra-analytics')) == 0,
            timeout=120, interval=10
        )

        show_step(4)
        waiters.wait(
            lambda: len(
                tf_manager.api.pods.list_starts_with('tf-zookeeper-nal')) == 0,
            timeout=60, interval=10
        )

        show_step(5)
        waiters.wait(
            lambda: len(
                tf_manager.api.pods.list_starts_with('tf-kafka')) == 0,
            timeout=60, interval=10
        )

        show_step(6)
        pvcs = tf_manager.api.pvolumeclaims.list(
            label_selector="app=tf-cassandra-analytics")
        pvcs.extend(tf_manager.api.pvolumeclaims.list(
            label_selector="app=tf-zookeeper-nal"))
        pvcs.extend(tf_manager.api.pvolumeclaims.list(
            label_selector="app=tf-kafka"))
        for pvc in pvcs:
            LOG.info(f"Delete pvc {pvc.name}")
            pvc.delete()

        show_step(7)
        pods = tf_manager.get_vrouter_pods()
        for pod in pods:
            LOG.info(f"Delete pod {pod.name}")
            pod.delete()

        show_step(8)
        tf_manager.wait_tf_controllers_healthy(timeout=360)
        tf_manager.wait_tfoperator_healthy(timeout=120)

        show_step(9)
        assert tf_manager.run_pytest() == 'Succeeded'


class TestTFTools(object):
    """Check enable/disable TF tools"""

    @pytest.fixture(scope='class', autouse=True)
    def restore_tfoperator_cr(self, tf_manager):
        LOG.info("Backup TFOperator specs")
        tfoperator = tf_manager.tfoperator()
        spec_backup = tfoperator.data['spec']
        yield
        LOG.info("Restore TFOperator specs")
        tfoperator.patch({"spec": spec_backup})
        tf_manager.wait_tf_controllers_healthy(timeout=180)
        tf_manager.wait_tfoperator_healthy(timeout=120)

    def test_contrail_tools(self, show_step, tf_manager):
        """ Check TF contrail tools

            Scenario:
                1. Enable contrail-tools
                2. Check contrail-tool pods
                3. Check contrail-tool cli
                4. Disable contrail-tools
                5. Check pods are deleted
        """
        pod_prefix = 'tf-tool-ctools'
        show_step(1)
        if tf_manager.is_ctools_enabled():
            LOG.info("TF contrail-tools is enabled. Disable it.")
            tf_manager.enable_ctools(state=False)
            waiters.wait(
                lambda: len(tf_manager.api.pods.list_starts_with(
                    pod_prefix)) == 0,
                timeout=150, interval=10
            )
        tf_manager.enable_ctools(state=True)

        show_step(2)
        vrouter_nodes = tf_manager.api.nodes.list(
            namespace=tf_manager.tf_namespace,
            label_selector='tfvrouter=enabled'
        )
        LOG.info(f"Amount of vrouter nodes is {len(vrouter_nodes)}")
        waiters.wait(
            lambda: len(tf_manager.api.pods.list_starts_with(
                pod_prefix)) == len(vrouter_nodes),
            timeout=150, interval=10
        )

        show_step(3)
        ctools_pods = tf_manager.api.pods.list(
            namespace=tf_manager.tf_namespace,
            name_prefix=pod_prefix)
        # Calling exec and check response code
        cmd = "vif --list"
        out = ctools_pods[0].exec(['/bin/sh', '-c', cmd])
        LOG.info(f"Command output:\n{out}")
        assert ("Vrouter Interface Table" in out), "Check vRouter cli."

        show_step(4)
        tf_manager.enable_ctools(state=False)

        show_step(5)
        waiters.wait(
            lambda: len(tf_manager.api.pods.list_starts_with(
                pod_prefix)) == 0,
            timeout=150, interval=10
        )

    def test_tf_cli(self, show_step, tf_manager):
        """ Check TF cli

            Scenario:
                1. Enable tf-cli tool
                2. Check tf-cli pods
                3. Check tf-cli
                4. Disable tf-cli tool
                5. Check pods are deleted
        """
        pod_prefix = 'tf-tool-cli'
        show_step(1)
        if tf_manager.is_tf_cli_enabled():
            LOG.info("TF cli is enabled. Disable it.")
            tf_manager.enable_tf_cli(state=False)
            waiters.wait(
                lambda: len(tf_manager.api.pods.list_starts_with(
                    pod_prefix)) == 0,
                timeout=150, interval=10
            )

        tf_manager.enable_tf_cli(state=True)

        show_step(2)
        waiters.wait(
            lambda: len(tf_manager.api.pods.list_starts_with(
                pod_prefix)) == 1,
            timeout=150, interval=10
        )

        show_step(3)
        ctools_pods = tf_manager.api.pods.list(
            namespace=tf_manager.tf_namespace,
            name_prefix=pod_prefix)
        # Calling exec and check response code
        cmd = "tf-ist-cli vr status"
        out = ctools_pods[0].exec(['/bin/sh', '-c', cmd])
        LOG.info(f"Command output:\n{out}")
        assert ("module_id: contrail-vrouter-agent" in out), "Check TF cli."

        show_step(4)
        tf_manager.enable_tf_cli(state=False)

        show_step(5)
        waiters.wait(
            lambda: len(tf_manager.api.pods.list_starts_with(
                pod_prefix)) == 0,
            timeout=150, interval=10
        )
